(******************************************************************************
tTabSet95 v2.3

Derived from
    tCustomControl

Properties
    AutoSize : Calculate tabs size & position automatically.
    BackColor : Background color for Tabs.
    ColorSelected : Color for selected Tab.
    UnColorSelected : Color for unselected Tab.
    MainHintIsZero : If true, first hint in list is main hint.
    Notebook : Notebook associed to these tabs.
    NotebookColor : If true, colors are the same that associed notebook.
    NotebookFont : If true, font is the same that associed notebook.
    Orientation : Orientation of tabset - Top, Left, Right or Bottom
    Style : Style of tabs (Standard or OwnerDraw).
    Tabs : List of Tabs in this TabSet.
    TabHeight : Height of Tabs when Autosize is false.
    TabHints : Hints for each Tab.
    TabIndex : Active tab.
    TabMargin : Margin before first Tab.
    TabsJustified : Set if tabs must be justified (use full width).
    TabSpacing : Space between each Tab.

Inherited properties
    Align, DragMode, DragCursor, Font, Height, Hint, ParentShowHint, ShowHint,
    TabOrder, TabStop, Visible, Width

Events
    BeforeChange
    AfterChange : no comment...

Inherited Events
    OnClick, OnDragDrop, OnDragOver, OnEndDrag, OnEnter, OnExit,
    OnKeyDown, OnKeyPress, OnKeyUp, OnMouseDown, OnMouseMove, OnMouseUp;

********************************************************************************
tTabbedNotebook95

Derived from
    tCustomControl

Properties
    MainHintIsZero : If true, first hint in list is main hint.
    Orientation : Orientation & position of tabset - Top, Left, Right or Bottom
    TabHeight : Height of Tabs when Autosize is false.
    TabHints : Hints for each Tab.
    TabFrontSize : Difference (in pixels) between normal tabs & active tab.
    TabSpacing : Space between each Tab.

********************************************************************************
Author name=BOURMAD Mehdi
Author E-mail=bourmad@multimania.com
Author URL=www.multimania.com/bourmad
******************************************************************************)

{If you have problems with these lines, remove them and mail me
 which compiler you have. I will correct it for future versions}
{$IFDEF WIN32} {Only for 32 bits compilers}
  {$IFNDEF VER90} {Doesn't works with Delphi 2}
    {$IFNDEF VER100} {Doesn't works with Delphi 3}
      {$ObjExportAll On} {export all objects (for C++Builder 3.0)}
    {$ENDIF}
  {$ENDIF}
{$ENDIF}

unit Tabs95;

interface

uses
  WinProcs, WinTypes, Classes, Controls, DsgnIntf, ExtCtrls, Forms, Graphics,
  Messages, SysUtils;

type
  TOrientation = (toBottom, toRight, toTop, toLeft);
  TTabStyle = (tsStandard,tsOwnerDraw);
  TTabChangeEvent = procedure (Sender: TObject; NewTab: Integer;
                               var AllowChange: Boolean) of object;
  pTabInfo = ^tTabInfo;
  tTabInfo = record
               Pos : tRect;
               Line,
               TextWidth,
               TextHeight : integer;
             end;

  {*********** tAboutMeProperty ************}
  TAboutMeProperty = class(TPropertyEditor)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
  end;

  {************** tTabSet95 ****************}
  tTabSet95 = class(TCustomControl)
  private
    { Stuff that is published }
    FAbout : TAboutMeProperty;
    FColorUnSelected,
    FColorSelected,
    FColorBack : TColor;
    FAutoSize : boolean;
    FChangeEvent : TTabChangeEvent;
    FMainHintisZero : boolean;
    FNbLines : integer;
    FNoteBook : TNoteBook;
    FNoteColor : boolean;
    FNoteFont : boolean;
    FOrientation : TOrientation;
    FStyle : TTabStyle;
    FTabList : TStrings;
    FTabFrontSize : integer;
    FTabHints : TStrings;
    FTabHeight : integer;
    FTabsJustified : Boolean;
    FTabSpacing : integer;
    FTabIndex : integer;
{    FTabFont : TFont;}
{    FOnClick: TNotifyEvent;}
    { stuff that is component internal }
    AllowChange : boolean;
    FTabDrawRects : tList;
  protected
    { overridden parent procedures }
    procedure Click; override;
    procedure CreateParams(var Params : TCreateParams); override;
    procedure MouseMove(Shift : TShiftState; X, Y : integer); override;
    procedure Notification(AComponent: TComponent;
                           Operation: TOperation); override;
    procedure Paint; override;
    { component specific procedures }
    procedure SetColorBack(Value : TColor);
    procedure SetColorSel(Value : TColor);
    procedure SetColorUnSel(Value : TColor);
    function LengthExt (aStr : String ) : integer;
    procedure CalculateTabs;
    procedure JustifyLines;
    procedure AdjustSize; {$IFDEF VER120}override;{$ENDIF}
    procedure BringToFront;
    procedure DrawTextHV (S : String; aRect: TRect);
    procedure DrawBorder (aRect : tRect; BackColor : tColor);
    function GetCorrectRect (TabInfo : tTabInfo) : tRect;
    procedure DrawTabs (TabCnt : Integer; RDraw : pRect);
    procedure SetAutoSize(Value : boolean);
    procedure SetMainHintIsZero(Value : boolean);
    procedure SetNoteBook(NB : TNoteBook);
    procedure SetNoteColor(Value : boolean);
    procedure SetNoteFont(Value : boolean);
    procedure SetOrientation(Value : TOrientation);
    procedure SetStyle(Value : TTabStyle);
    procedure SetTabsJustified(Value : boolean);
    procedure SetTabSpacing(Value : integer);
    procedure SetTabHeight(Value : integer);
    procedure SetTabFrontSize(Value : integer);
    procedure SetTabHints(List : TStrings);
    procedure SetTabList(List : TStrings);
    procedure SetTabIndex(Value : integer);
    { message handlers }
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure CMDialogChar(var Msg : TCMDialogChar); message cm_DialogChar;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    { Public declarations }
  published
    property About: TAboutMeProperty read FAbout write FAbout;
    property ColorBack : TColor read FColorBack
                                write SetColorBack default clGray;
    property ColorSelected : TColor read FColorSelected
                                    write SetColorSel default clTeal;
    property ColorUnSelected : TColor read FColorUnSelected
                                      write SetColorUnSel default clSilver;
    property AutoSize : boolean read FAutoSize write SetAutoSize default TRUE;
    property MainHintIsZero : boolean read FMainHintIsZero
                                      write SetMainHintIsZero default TRUE;
    property Notebook : TNoteBook read FNotebook write SetNotebook;
    property NotebookColor : boolean read FNoteColor
                                     write SetNoteColor default TRUE;
    property NotebookFont : boolean read FNoteFont
                                    write SetNoteFont default TRUE;
    property OnChange : TTabChangeEvent read FChangeEvent write FChangeEvent;
    property Orientation : TOrientation read FOrientation
                                        write SetOrientation default toTop;
    property Style : TTabStyle read FStyle write SetStyle default tsStandard;
    property Tabs : TStrings read FTabList write SetTabList;
    property TabFrontSize  : integer read FTabFrontSize write SetTabFrontSize default 2;
    property TabHeight : integer read FTabHeight write SetTabHeight default 25;
    property TabHints : TStrings read FTabHints write SetTabHints;
    property TabIndex : integer read FTabIndex write SetTabIndex default 0;
    property TabsJustified : boolean read FTabsJustified write SetTabsJustified default true;
    property TabSpacing : integer read FTabSpacing write SetTabSpacing default 0;
    {inherited properties }
    property Align;
    property DragMode;
    property DragCursor;
    property Font;
    property Height default 25; {TabHeight+2}
    property Hint;
    property ParentShowHint;
    property ShowHint;
    property TabOrder;
    property TabStop default TRUE;
    property Visible default TRUE;
    property Width default 200;
    property OnClick{: TNotifyEvent read FOnClick write FOnClick};  
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
{    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;}
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

  {************** tTabbedNotebook95 ****************}
  tTabbedNotebook95 = class(TCustomControl)
  private
    FAbout : TAboutMeProperty;
    Notebook : tNotebook;
    TabSet : tTabSet95;
  protected
    function GetActivePage : String;
    procedure SetActivePage (Value : String);
    function GetMainHintIsZero : boolean;
    procedure SetMainHintIsZero(Value : boolean);
    function GetOrientation : TOrientation;
    procedure SetOrientation (Value : TOrientation);
    function GetPageIndex : Integer;
    procedure SetPageIndex(Value: Integer);
    function GetPages : tStrings;
    procedure SetPages (Value : tStrings);
    function GetTabHeight : integer;
    procedure SetTabHeight(Value : integer);
    function GetTabHints : TStrings;
    procedure SetTabHints(List : TStrings);
    function GetTabSpacing : integer;
    procedure SetTabSpacing(Value : integer);

{   procedure CMTabFontChanged(var Message:TMessage);message CM_TABFONTCHANGED;
    procedure ReadState(Reader: TReader); override;
    procedure WriteComponents(Writer: TWriter); override;}
(*    procedure Loaded; override;
    procedure SetName(const NewName: TComponentName); override;
    procedure Notification(AComponent: TComponent;
                           Operation: TOperation); override;*)
    procedure Paint; override;
    procedure AlignControls(AControl: TControl; var Rect: TRect); override;
{    procedure CreateParams(var Params: TCreateParams); override;}
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function GetIndexForPage(const PageName: String): Integer;
{    property TopFont: TFont;
    procedure TabFontChanged(Sender: TObject);}
  published
    property About: TAboutMeProperty read FAbout write FAbout;
    property MainHintIsZero : boolean read GetMainHintIsZero
                                      write SetMainHintIsZero default TRUE;
    property Orientation : TOrientation read GetOrientation
                                        write SetOrientation default toTop;
    property TabHeight : integer read GetTabHeight write SetTabHeight default 25;
    property TabHints : TStrings read GetTabHints write SetTabHints;
    property TabSpacing : integer read GetTabSpacing write SetTabSpacing default 0;

    property ActivePage: string read GetActivePage write SetActivePage;
    property Align;
    property Enabled;
    property PageIndex: Integer read GetPageIndex write SetPageIndex default 0;
    property Pages: TStrings read GetPages write SetPages;
    property Font;
{    property TabsPerRow: Integer default 3;
    property TabFont: TFont;}
    property ParentShowHint;
    property ShowHint;
    property TabOrder;
    property TabStop default True;
    property Width default 300;
    property Height default 250;
    property Visible;
    property OnClick;
{    property OnChange: TPageChangeEvent;}
    property OnEnter;
    property OnExit;
  end;

procedure Register;

implementation

{$IFNDEF WIN32}
  {$R *.D16}
{$ELSE}
  {$R *.D32}
{$ENDIF}

procedure Register;
begin
(*  RegisterPropertyEditor (TypeInfo(TAboutMeProperty), tTabbedNotebook95,
                          'ABOUT', TAboutMeProperty);
  RegisterComponents ('BmSuite', [tTabSet95, tTabbedNotebook95]);*)

{i've removed tTabbedNotebook95's registration because this comp doesn't work !}
  RegisterPropertyEditor (TypeInfo(TAboutMeProperty), tTabSet95,
                          'ABOUT', TAboutMeProperty);
  RegisterComponents ('BmSuite', [tTabSet95]);
end;

{************************* tAboutMeProperty component *************************}
procedure TAboutMeProperty.Edit;
begin
  Application.MessageBox ('tTabSet95 & tTabbedNotebook95 components v2.3.'
                         +#13#10'These components are freeware.'
                         +#13#10' 1997-1999 BOURMAD Mehdi'
                         +#13#10'http://www.multimania.com/bourmad'
                         +#13#10'mailto:bourmad@multimania.com',
                         'About',
                         MB_OK+ MB_ICONINFORMATION);
end;

function TAboutMeProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paMultiSelect, paDialog, paReadOnly];
end;

function TAboutMeProperty.GetValue: string;
begin
  Result := '(about)';
end;

{************** tTabSet95 ****************}
constructor tTabSet95.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls];
  Width := 200;
  FOrientation := toTop;
  FTabFrontSize := 2;
  FTabsJustified := true;
  FTabSpacing := 0;
  FTabHeight := 25;
  Height := FTabHeight+2;
  FTabIndex := 0;
  FStyle := tsStandard;
  TabStop := TRUE;
  FAutoSize := TRUE;
  FColorBack := clGray;
  FColorSelected := clTeal;
  FColorUnSelected := clSilver;
  FNoteColor := True;
  FNoteFont := True;
  FNotebook := nil;
  FTabDrawRects := tList.Create;
  FTabList := TStringList.Create;
  FTabHints := TStringList.Create;
  FMainHintIsZero := TRUE;
  with Font do
  begin
    Name := 'MS Sans Serif';
    Height := -10;
    Color := clBlack;
    Style := [];
  end;
end;

destructor tTabSet95.Destroy;
begin
  FTabDrawRects.Free;
  FTabList.Free;
  FTabHints.Free;
  if FNotebook <> nil
    then FNotebook := nil;
  inherited Destroy;
end;

procedure tTabSet95.CreateParams(var Params : TCreateParams);
begin
  inherited CreateParams(Params);
end;

procedure tTabSet95.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FNotebook) and (Operation = opRemove)
    then FNotebook := nil
end;

procedure tTabSet95.SetOrientation(Value : TOrientation);
begin
  if FOrientation <> Value
  then begin
    FOrientation := Value;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetColorBack(Value : TColor);
begin
  if FColorBack <> Value
  then begin
    FColorBack := Value;
    FNoteColor := False;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetColorSel(Value : TColor);
begin
  if FColorSelected <> Value
  then begin
    FColorSelected := Value;
    FNoteColor := False;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetColorUnSel(Value : TColor);
begin
  if FColorUnSelected <> Value
  then begin
    FColorUnSelected := Value;
    FNoteColor := False;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetTabFrontSize(Value : integer);
begin
  if FTabFrontSize <> Value
  then begin
    if Value < 0 then Value := 0;
    FTabFrontSize := Value;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetTabHeight(Value : integer);
begin
  if FTabHeight <> Value
  then begin
    if Value < 0 then Value := 0;
    FTabHeight := Value;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetTabsJustified(Value : boolean);
begin
  if FTabsJustified <> Value
  then begin
    FTabsJustified := Value;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetTabSpacing(Value : integer);
begin
  if FTabSpacing <> Value
  then begin
    FTabSpacing := Value;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetStyle(Value : TTabStyle);
begin
  if FStyle <> Value
  then begin
    FStyle := Value;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetAutoSize(Value : boolean);
begin
  if FAutoSize <> Value
  then begin
    FAutoSize := Value;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetTabIndex(Value : integer);
begin
  if FTabIndex <> Value
  then begin
    AllowChange := TRUE;
    if assigned(FChangeEvent)
      then FChangeEvent(Self,Value,AllowChange);
    if AllowChange
    then begin
      if Value < 0
        then Value := 0;
      if Value >= FTabList.Count
        then Value := FTabList.Count-1;
      FTabIndex := Value;
      AdjustSize;
      Invalidate;
    end;
  end;
end;

procedure tTabSet95.SetTabList(List : TStrings);
begin
  if FTabList <> nil
  then begin
    FTabList.Clear;
  end;
  if List <> nil
  then begin
    FTabList.Assign(List);
  end;
  AdjustSize;
  Invalidate;
end;

procedure tTabSet95.SetTabHints(List : TStrings);

begin
  if FTabHints <> nil
  then begin
    FTabHints.Clear;
    if List <> nil
      then FTabHints.Assign(List);
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetMainHintIsZero(Value : boolean);
begin
  if FMainHintIsZero <> Value
  then begin
    FMainHintIsZero := Value;
    AdjustSize;
    Invalidate;
  end;
end;

procedure tTabSet95.SetNoteBook(NB : TNoteBook);
begin
  FNoteBook := NB;
  FTabList.Clear;
  if NB <> nil then
  begin
    FTabList.Assign(NB.Pages);
    if FNoteColor
    then begin
      FColorSelected := FNoteBook.Color;
      FColorUnSelected := FNoteBook.Color;
      FColorBack := FNoteBook.Color;
    end;
    if FNoteFont
    then begin
      Font := FNoteBook.Font;
    end;
  end;
  AdjustSize;
  Invalidate;
end;

procedure tTabSet95.SetNoteColor(Value : boolean);
begin
  FNoteColor := Value;
  if FNoteColor and (FNoteBook <> nil)
  then begin
    FColorSelected := FNoteBook.Color;
    FColorUnSelected := FNoteBook.Color;
    FColorBack := FNoteBook.Color;
  end;
  AdjustSize;
  Invalidate;
end;

procedure tTabSet95.SetNoteFont(Value : boolean);
begin
  FNoteFont := Value;
  if FNoteFont and (FNoteBook <> nil)
  then begin
    Font := FNoteBook.Font;
  end;
  AdjustSize;
  Invalidate;
end;

procedure tTabSet95.MouseMove(Shift : TShiftState; X, Y : integer);
var
  Found : boolean;
  i : integer;
  aTab : pTabInfo;
begin
  inherited MouseMove(Shift,X,Y);
  Hint := '';
  Found := FALSE;
  if (FTabList.Count > 0) and (FTabHints.Count > 0) and ShowHint
  then begin
    for i := 0 to FTabHints.Count - 1 do
    begin
{      if PtInRect(tTabInfo(FTabDrawRects.Items[i]^).Pos, Point(X,Y))}
      aTab := FTabDrawRects.Items[i];
      if PtInRect(GetCorrectRect(aTab^), Point(X,Y))
      then begin
        if (FTabHints.Count > 0) and (length(FTabHints[i]) > 0)
        then begin
          if FMainHintIsZero then
            Hint := FTabHints[i + 1]
          else
            Hint := FTabHints[i];
          Found := TRUE;
          break;
        end;
      end;
    end;
    if not Found and FMainHintIsZero
      then Hint := FTabHints[0];
  end;
end;


procedure tTabSet95.WMSetFocus(var Message: TWMSetFocus);
begin
  inherited;
  Invalidate;
end;

procedure tTabSet95.WMSize(var Message: TWMSize);
begin
  inherited;
  AdjustSize;
  invalidate;
end;

procedure tTabSet95.WMKillFocus(var Message: TWMKillFocus);
begin
  inherited;
  Invalidate;
end;

procedure tTabSet95.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;
end;

procedure tTabSet95.KeyDown(var Key: Word; Shift: TShiftState);
var
  FOnClick : tNotifyEvent;
begin
  case Key of
    VK_RIGHT, VK_DOWN:
      begin
        if FTabIndex >= (FTabList.Count-1) then SetTabIndex(0)
        else SetTabIndex(FTabIndex + 1);
        FOnClick := OnClick;
        if Assigned (FOnClick)
          then OnClick(Self);
      end;
    VK_LEFT, VK_UP:
      begin
        if FTabIndex > 0 then SetTabIndex(FTabIndex - 1)
        else SetTabIndex(FTabList.Count - 1);
        FOnClick := OnClick;
        if Assigned (FOnClick)
          then OnClick(Self);
      end;
  end;
end;

{ This provides entry to accel chars from the main form keyboard loop }
procedure tTabSet95.CMDialogChar( var Msg : TCMDialogChar );
var
  i : integer;
begin
  with Msg do
  begin
    if FTabList.Count > 0
    then begin
      for i := 0 to FTabList.Count - 1 do
      begin
        if IsAccel(CharCode,FTabList[i])
        then begin
          if assigned(FChangeEvent)
            then FChangeEvent(Self,i,AllowChange);
          if AllowChange
          then begin
            FTabIndex := i;
            SetFocus;
            Invalidate;
          end;
          Result := 1;  { say we handled it, even if AllowChange is false }
          break;
        end;
      end;
    end
    else
      inherited;
  end;
end;

procedure tTabSet95.Click;
var
  MP,TP : TPoint;
  i : integer;
  aTab : pTabInfo;
  Found : boolean;
begin
  GetCursorPos(MP);
  TP := ScreenToClient(MP);
  Found := false;
  if FTabList.Count > 0 then
  begin
    i := 0;
    while (i<=FTabList.Count - 1) and not Found do
    begin
      aTab := FTabDrawRects.Items[i];
      if PtInRect(GetCorrectRect(aTab^), TP)
      then begin
        SetTabIndex(i);
        SetFocus;
        Found := true;
      end;
      inc(i);
    end;
    Invalidate;
  end;
  inherited Click;
end;

function tTabSet95.LengthExt (aStr : String ) : integer;
begin
  Result := Length(aStr);
  if Pos ('&', aStr)<>0
    then dec (Result);
end;

procedure tTabSet95.CalculateTabs;
var
  aLine,
  aSize,
  TabCnt,
  Cx, Cy,
  CharW,CharH : integer;
  TabCaption : string;
  TabInfo : pTabInfo;
begin
  Canvas.Font := Self.Font;
  If fStyle=tsStandard
    then FTabHeight := Muldiv (Canvas.TextHeight('W'),150,100);
   {else, keep value defined by user}

  CharW := Canvas.TextWidth('W');
  CharH := Canvas.TextHeight('W');

  FTabDrawRects.Clear;
  {starting point for tabs}
  Cx := 0; Cy := 0;
  case FOrientation of
       toTop,
    toBottom : begin
                 Cx := FTabFrontSize;
                 Cy := FTabHeight;
               end;
      toLeft,
     toRight : begin
                 Cx := FTabHeight;
                 Cy := FTabFrontSize;
               end;
  end;

  aLine := 1;
  FNbLines := 0;
  {place each tabs...}
  for TabCnt := 0 to FTabList.Count - 1 do
  begin
    TabInfo := new (pTabInfo);
    TabCaption := FTabList[TabCnt];
    {size of tab's caption, rect for drawing and for showing activated this tab}
    case FOrientation of
         toTop,
      toBottom : begin
                   TabInfo^.TextWidth := Canvas.TextWidth(TabCaption);
                   if Pos ('&', TabCaption)<>0
                     then dec (TabInfo^.TextWidth, CharW);
                   if TabInfo^.TextWidth<=CharW
                     then inc (TabInfo^.TextWidth, CharW div 2);
                   TabInfo^.TextHeight:=CharH;

                   aSize := MulDiv(TabInfo^.TextWidth,150,100)+4;
                   if Cx+aSize>Width
                   then begin
                     aLine := aLine + 1;
                     Cx := FTabFrontSize;
                   end;
                   TabInfo^.Pos := Rect(Cx, 0, Cx+aSize, FTabHeight);
                 end;
        toLeft,
       toRight : begin
                   TabInfo^.TextWidth := CharH*LengthExt(TabCaption);
                   if TabInfo^.TextWidth<=CharH
                     then inc (TabInfo^.TextWidth, CharH div 2);
                   TabInfo^.TextHeight:=CharW;

                   aSize := TabInfo^.TextWidth + CharH;
                   if Cy+aSize>Height
                   then begin
                     aLine := aLine + 1;
                     Cy := FTabFrontSize;
                   end;
                   TabInfo^.Pos := Rect(0, Cy, FTabHeight, Cy+aSize);
                 end;
    end;
    {tab's line}
    TabInfo^.Line := aLine;
    if aLine > FNbLines
      then FNbLines := aLine;
    {add tab to tabs list}
    FTabDrawRects.Add (TabInfo);
    {starting point for next tab}
    case FOrientation of
         toTop,
      toBottom : begin
                   inc (Cx,TabInfo^.Pos.Right-TabInfo^.Pos.Left + 1 + FTabSpacing);
                 end;
        toLeft,
       toRight : begin
                   inc (Cy,TabInfo^.Pos.Bottom-TabInfo^.Pos.Top + 1 + FTabSpacing);
                 end;
    end;
  end; {end for}
end;

procedure tTabSet95.JustifyLines;
var
  aWidth, aSize,
  Cx, Cy,
  aNbChar, aTabChar,
  Tab1, Tab2,
  NumLine, TabCnt : integer;
begin
  {recalculate width for each tabs, for use full tabset width}
  for NumLine := 1 to FNbLines do
  begin
    Tab1 := -1;
    Tab2 := -1;
    aNbChar := 0;
    for TabCnt := 0 to FTabList.Count - 1 do
    begin {get start and end (tab) of this line, and its width (in char)}
      if (tTabInfo(FTabDrawRects.Items[TabCnt]^).Line=NumLine)
      then begin
        if Tab1 = -1
        then begin
          Tab1 := TabCnt;
        end;
        Tab2 := TabCnt;
        aNbChar := aNbChar + LengthExt(FTabList[TabCnt]);
      end;
    end;
    Cx := 0; Cy := 0;
    aWidth := 0;
    case FOrientation of
         toTop,
      toBottom : begin
                   Cx := FTabFrontSize;
                   Cy := FTabHeight;
                   aWidth := Width-FTabFrontSize*2;
                 end;
        toLeft,
       toRight : begin
                   Cx := FTabHeight;
                   Cy := FTabFrontSize;
                   aWidth := Height-FTabFrontSize*2;
                 end;
    end;
    for TabCnt := Tab1 to Tab2 do
    begin {set new width for this line's tabs}
      case FOrientation of
           toTop,
        toBottom : begin
                     aTabChar := LengthExt(FTabList[TabCnt]);
                     aSize := Trunc(aWidth/aNbChar*aTabChar);
                     tTabInfo(FTabDrawRects.Items[TabCnt]^).Pos.Left := Cx;
                     tTabInfo(FTabDrawRects.Items[TabCnt]^).Pos.Right := Cx+aSize-1;
                     inc (Cx, aSize + FTabSpacing);
                     aWidth := aWidth - aSize - FTabSpacing;
                     aNbChar := aNbChar - aTabChar;
                   end;
          toLeft,
         toRight : begin
                     aTabChar := LengthExt(FTabList[TabCnt]);
                     aSize := Trunc(aWidth/aNbChar*aTabChar);
                     tTabInfo(FTabDrawRects.Items[TabCnt]^).Pos.Top := Cy;
                     tTabInfo(FTabDrawRects.Items[TabCnt]^).Pos.Bottom := Cy+aSize-1;
                     inc (Cy, aSize + FTabSpacing);
                     aWidth := aWidth - aSize - FTabSpacing;
                     aNbChar := aNbChar - aTabChar;
                  end;
      end; { Orientation case }
    end;
  end;
end;

procedure tTabSet95.AdjustSize;
begin
  {$IFDEF WIN32}
  inherited;
  {$ENDIF}
  {first, calculate fixed size for TabSet}
  if FAutoSize and (FNotebook <> nil) then
  begin
    case FOrientation of
         toTop,
      toBottom : begin
                   Width := FNoteBook.Width;
                   Left := FNoteBook.Left;
                 end;
        toLeft,
       toRight : begin
                   Height := FNotebook.Height;
                   Top := FNoteBook.Top;
                end;
    end; { Orientation case }
  end;
  if FTabList.Count > 0
  then begin
    {calculate position (line) for each tabs}
    CalculateTabs;
    {recalculate width for each tabs, for use full tabset width}
    if FTabsJustified
      then JustifyLines;
  end;
  {then, calculate fixed size for TabSet}
  if FAutoSize and (FNotebook <> nil) then
  begin
    {calculate general size for TabSet}
    case FOrientation of
         toTop : begin
                   Height := FTabHeight*FNbLines + FTabFrontSize+1;
                   Top := FNoteBook.Top - Height;
                 end;
      toBottom : begin
                   Height := FTabHeight*FNbLines + FTabFrontSize+1;
                   Top := FNoteBook.Top + FNoteBook.Height;
                 end;
        toLeft : begin
                   Width := FTabHeight*FNbLines + FTabFrontSize+1;
                   Left := FNotebook.Left - Width;
                 end;
       toRight : begin
                   Width := FTabHeight*FNbLines + FTabFrontSize+1;
                   Left := FNotebook.Left + FNotebook.Width;
                end;
    end; { Orientation case }
  end;
end;

procedure tTabSet95.BringToFront;
var
  Delta,
  NewLine,
  TabCnt : integer;
  aTab : pTabInfo;
begin
  Delta := tTabInfo(FTabDrawRects.Items[FTabIndex]^).Line-1;
  if Delta<>0
  then begin
    for TabCnt:=0 to FTabList.Count - 1 do
    begin
      aTab := FTabDrawRects.Items[TabCnt];
      NewLine := aTab^.Line-Delta;
      if NewLine<=0
        then NewLine := FNbLines+NewLine;
      aTab^.Line := NewLine;
    end;
  end;
end;

procedure tTabSet95.Paint;
var
  RDraw : tRect;
  TabCnt : integer;
  StartPnt, EndPnt : TPoint;
begin
  Canvas.Brush.Color := FColorBack;
  Canvas.Pen.Style := psSolid;
  Canvas.Pen.Width := 1;
  {clear client region (all tabset)}
  Canvas.FillRect(ClientRect);
  case FOrientation of
       toTop : begin
                 StartPnt := Point(0, Height-1);
                 EndPnt := Point(Width, Height-1);
                 Canvas.Pen.Color := clBtnHighlight;
               end;
    toBottom : begin
                 StartPnt := Point(0, 0);
                 EndPnt := Point(Width, 0);
                 Canvas.Pen.Color := clBtnShadow;
               end;
      toLeft : begin
                 StartPnt := Point(Width-1, 0);
                 EndPnt := Point(Width-1, Height);
                 Canvas.Pen.Color := clBtnHighlight;
               end;
     toRight : begin
                 StartPnt := Point(0, 0);
                 EndPnt := Point(0, Height);
                 Canvas.Pen.Color := clBtnShadow;
               end;
  end;
  {draw continuous line of TabSet}
  Canvas.MoveTo(StartPnt.X, StartPnt.Y);
  Canvas.LineTo(EndPnt.X, EndPnt.Y);

  if FTabList.Count > 0
  then begin
    {Bring to front, line with active tab}
    BringToFront;
    {draw all tabs}
    for TabCnt := 0 to FTabList.Count - 1 do
      DrawTabs (TabCnt, nil); {nil : use FTabDrawRects for drawing}

    {draw focused tab, bigger than others}
    if FTabList.Count > 0 then
    begin
      RDraw := GetCorrectRect(tTabInfo(FTabDrawRects.Items[FTabIndex]^));
      {erase normal border for this tab}
      Canvas.Pen.Color := FColorBack;
      Canvas.Brush.Color := FColorBack;
      case FOrientation of
           toTop : Canvas.Rectangle (RDraw.Left-1, RDraw.Top,
                                     RDraw.Right+1, RDraw.Bottom);
        toBottom : Canvas.Rectangle (RDraw.Left-1, RDraw.Top,
                                     RDraw.Right+1, RDraw.Bottom+1);
          toLeft : Canvas.Rectangle (RDraw.Left, RDraw.Top-1,
                                     RDraw.Right+1, RDraw.Bottom+1);
         toRight : Canvas.Rectangle (RDraw.Left, RDraw.Top-1,
                                     RDraw.Right+1, RDraw.Bottom+1);
      end;
      {draw bigger border...}
      InflateRect (RDraw, FTabFrontSize, FTabFrontSize);
      DrawTabs (FTabIndex, @RDraw); {use RDraw instead of FTabDrawRects for drawing}
      {draw focused rectangle}
      InflateRect(RDraw, -FTabFrontSize, -FTabFrontSize);
      if Focused
        then Canvas.DrawFocusRect(RDraw);
    end;
    {change assigned notebook page}
    if Assigned(FNoteBook)
      then FNotebook.PageIndex := FTabIndex;
  end;
end;

procedure tTabSet95.DrawTextHV (S : String; aRect: TRect);
var
  Buf         : Array[0..255] of char;
  RectText    : tRect;
  aStr        : String;
  CharH,
  CharTop,
  STextWidth,
  I           : Integer;
begin
  if FOrientation in [toLeft, toRight]
  then begin
    CharH := Canvas.TextHeight('W');
    STextWidth := CharH*LengthExt(S);
    if Pos ('&', S)<>0
      then dec (STextWidth, CharH);
    CharTop := aRect.Top+((aRect.Bottom-aRect.Top)-STextWidth) div 2;
    RectText := aRect;
    i := 1;
    while i<= Length(S) do
    begin
      aStr := Copy (S, I, 1);
      if aStr='&'
      then begin
        inc(i);
        aStr := aStr + Copy(S, I, 1);
      end;
      StrPCopy (Buf, aStr);
      RectText := Rect(RectText.Left, CharTop,
                       RectText.Right, CharTop + CharH);
      DrawText(Canvas.Handle, Buf, Length(aStr), RectText,
               DT_CENTER or DT_VCENTER or DT_SINGLELINE);
      inc (CharTop, CharH);
      inc(i);
    end;
  end
  else begin
    StrPCopy (Buf, S);
    DrawText (Canvas.Handle, Buf, length(S), aRect,
              DT_VCENTER or DT_CENTER or DT_SINGLELINE);
  end;
end;

procedure tTabSet95.DrawBorder (aRect : tRect; BackColor : tColor);
var
  Corner1, Corner1b,
  Corner2, Corner2b,
  Corner3, Corner3b, Corner3c,
  Corner4,  Corner4b : tPoint;
begin
  Canvas.Pen.Width := 1;
  {calculate tab's polygon points}
  case FOrientation of
    toTop : begin
      Corner1 := Point(aRect.Left, aRect.Bottom);
      Corner1b := Point (Corner1.X, Corner1.Y-1);
      Corner2 := Point(Corner1.X, aRect.Top+2);
      Corner2b := Point (Corner2.X+2, Corner2.Y-2);
      Corner3 := Point (aRect.Right-2, Corner2b.Y);
      Corner3b := Point (Corner3.X+1, Corner3.Y+1);
      Corner3c := Point (Corner3b.X+1, Corner3b.Y+1);
      Corner4 := Point(Corner3c.X, Corner1.Y-1);
      Corner4b := Point (Corner4.X-1, Corner4.Y);
    end;
    toBottom : begin
      Corner1 := Point(aRect.Left,aRect.Top);
      Corner1b := Point (Corner1.X, Corner1.Y+1);
      Corner2 := Point(Corner1.X,aRect.Bottom-2);
      Corner2b := Point (Corner2.X+2, Corner2.Y+2);
      Corner3 := Point (aRect.Right-2, Corner2b.Y);
      Corner3b := Point (Corner3.X+1, Corner3.Y-1);
      Corner3c := Point (Corner3b.X+1, Corner3b.Y-1);
      Corner4 := Point(Corner3c.X, Corner1.Y+1);
      Corner4b := Point (Corner4.X-1, Corner4.Y);
    end;
    toLeft : begin
      Corner1 := Point(aRect.Right, aRect.Top);
      Corner1b := Point (Corner1.X-1, Corner1.Y);
      Corner2 := Point(aRect.Left+2, Corner1.Y);
      Corner2b := Point (Corner2.X-2, Corner2.Y+2);
      Corner3 := Point(Corner2b.X , aRect.Bottom-2);
      Corner3b := Point (Corner3.X+1, Corner3.Y+1);
      Corner3c := Point (Corner3b.X+1, Corner3b.Y+1);
      Corner4 := Point(Corner1.X-1, Corner3c.Y);
      Corner4b := Point (Corner4.X, Corner4.Y-1);
    end;
    toRight : begin
      Corner1 := Point(aRect.Left, aRect.Top);
      Corner1b := Point (Corner1.X+1, Corner1.Y);
      Corner2 := Point(aRect.Right-2, Corner1.Y);
      Corner2b := Point (Corner2.X+2, Corner2.Y+2);
      Corner3 := Point(Corner2b.X ,aRect.Bottom-2);
      Corner3b := Point (Corner3.X-1, Corner3.Y+1);
      Corner3c := Point (Corner3b.X-1, Corner3b.Y+1);
      Corner4 := Point(Corner1.X+1, Corner3c.Y);
      Corner4b := Point (Corner4.X, Corner4.Y-1);
    end;
  end;

  {draw polygon filling current tab with BackColor}
  Canvas.Brush.Color := BackColor;
  Canvas.Pen.Color := BackColor;
  Canvas.Polygon ([Corner1b, Corner2, Corner2b, Corner3, Corner3c, Corner4]);

  {with PolyLine, last pixel isn't drawn...}
  case FOrientation of
    toTop : begin
      Canvas.Pen.Color := clBtnShadow;
      Canvas.PolyLine ([Corner4b, Corner3b]);
      Canvas.Pen.Color := clBlack;
      Canvas.PolyLine ([Corner4, Corner3c, Corner3]);
      Canvas.Pen.Color := clBtnHighlight;
      Canvas.PolyLine ([Corner3, Corner2b, Corner2, Corner1]);
    end;
    toBottom : begin
      Canvas.Pen.Color := clBtnShadow;
      Canvas.PolyLine ([Corner4b, Corner3b]);
      Canvas.Pen.Color := clBlack;
      Canvas.PolyLine ([Corner4, Corner3c, Corner3, Corner2b]);
      Canvas.Pen.Color := clBtnHighlight;
      Canvas.PolyLine ([Corner2b, Corner2, Corner1]);
    end;
    toLeft : begin
      Canvas.Pen.Color := clBtnShadow;
      Canvas.PolyLine ([Corner4b, Corner3b]);
      Canvas.Pen.Color := clBlack;
      Canvas.PolyLine ([Corner4, Corner3c, Corner3]);
      Canvas.Pen.Color := clBtnHighlight;
      Canvas.PolyLine ([Corner3, Corner2b, Corner2, Corner1]);
    end;
    toRight : begin
      Canvas.Pen.Color := clBtnShadow;
      Canvas.PolyLine ([Corner4b, Corner3b]);
      Canvas.Pen.Color := clBlack;
      Canvas.PolyLine ([Corner4, Corner3c, Corner3, Corner2b]);
      Canvas.Pen.Color := clBtnHighlight;
      Canvas.PolyLine ([Corner2b, Corner2, Corner1]);
    end;
  end;
end;

function tTabSet95.GetCorrectRect (TabInfo : tTabInfo) : tRect;
begin
  Result := TabInfo.Pos;
  {rect for drawing and for showing activated this tab}
  case FOrientation of
       toTop : begin
                 Result.Top := Height-FTabHeight*TabInfo.Line-1;
                 Result.Bottom := Result.Top + Result.Bottom;
               end;
    toBottom : begin
                 Result.Top := FTabHeight*(TabInfo.Line-1);
                 Result.Bottom := Result.Top + Result.Bottom;
               end;
      toLeft : begin
                 Result.Left := Width-FTabHeight*TabInfo.Line-1;
                 Result.Right := Result.Left + Result.Right;
               end;
     toRight : begin
                 Result.Left := FTabHeight*(TabInfo.Line-1);
                 Result.Right := Result.Left + Result.Right;
               end;
  end;
end;

procedure tTabSet95.DrawTabs (TabCnt : Integer; RDraw : pRect);
var
  BackColor : tColor;
  Rect : tRect;
begin
  {background color of active tab}
  if TabCnt <> FTabIndex
    then BackColor := FColorUnSelected
    else BackColor := FColorSelected;

  {calculate and draw current}
  if assigned(RDraw)
    then Rect := RDraw^
    else Rect := GetCorrectRect(tTabInfo(FTabDrawRects.Items[TabCnt]^));
  DrawBorder (Rect, BackColor);

  If fStyle=tsStandard {style "standard" : needs to write text into tabs}
    then DrawTextHV (FTabList[TabCnt], Rect);
end;

{************** tTabbedNotebook95 *****************}
constructor tTabbedNotebook95.Create(AOwner: TComponent);
begin
  Inherited Create (aOwner);
{  Parent:=tWinControl(AOwner);}

  Width := 300;
  Height := 250;

  Notebook := tNotebook.Create(Self);
  Notebook.Name:='nb'+Self.Name;
  Notebook.Parent := Self;

  TabSet := tTabSet95.Create (Self);
  TabSet.Name:='ts'+Self.Name;
  TabSet.Parent := Self;

(*  Notebook := tNotebook.Create(Self);
  Notebook.Name:='nb'+Self.Name;
  Notebook.Parent := Self;

  TabSet := tTabSet95.Create (Self);
  TabSet.Name:='ts'+Self.Name;
  TabSet.Parent := Self;*)

{  ControlStyle := ControlStyle + [csAcceptsControls];}

{  TabSet.Notebook := NoteBook;}
{  TabSet.Tabs.Assign(Notebook.Pages);}
{  TabSet.Tabs := Notebook.Pages;}
{  TabSet.ColorSelected := NoteBook.Color;
  TabSet.ColorUnSelected := NoteBook.Color;
  TabSet.ColorBack := NoteBook.Color;}

{  messagebeep(0);}
end;

destructor tTabbedNotebook95.Destroy;
begin
  TabSet.Free;
  NoteBook.Free;
  inherited Destroy;
end;

(*procedure tTabbedNotebook95.SetName(const NewName: TComponentName);
begin
  inherited SetName (NewName);
{  if csDesigning in ComponentState
  then begin
    Notebook.Name := 'NB'+Name;
    TabSet.Name := 'TS'+Name;
  end;}
end;*)

(*procedure tTabbedNotebook95.Loaded;
begin
  inherited Loaded;
  Parent:=tWinControl(Owner);
  Notebook := tNotebook.Create(Self);
  Notebook.Name:='nb'+Self.Name;
  Notebook.Parent := Self;

  TabSet := tTabSet95.Create (Self);
  TabSet.Name:='ts'+Self.Name;
  TabSet.Parent := Self;

  TabSet.Notebook := NoteBook;
{  TabSet.Tabs.Assign(Notebook.Pages);}
  TabSet.Tabs := Notebook.Pages;
{  TabSet.ColorSelected := NoteBook.Color;
  TabSet.ColorUnSelected := NoteBook.Color;
  TabSet.ColorBack := NoteBook.Color;}
end;*)

(*procedure tTabbedNotebook95.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opInsert) and (aComponent.Owner=Self)
  then begin
  end;
end;*)

function tTabbedNotebook95.GetActivePage : String;
begin
  Result := Notebook.ActivePage;
end;

procedure tTabbedNotebook95.SetActivePage (Value : String);
begin
  Notebook.ActivePage := Value;
  TabSet.TabIndex := Notebook.PageIndex;
end;

function tTabbedNotebook95.GetMainHintIsZero : boolean;
begin
  Result := TabSet.MainHintIsZero;
end;

procedure tTabbedNotebook95.SetMainHintIsZero(Value : boolean);
begin
  TabSet.MainHintIsZero := Value;
end;

function tTabbedNotebook95.GetOrientation : TOrientation;
begin
  Result := TabSet.Orientation;
end;

procedure tTabbedNotebook95.SetOrientation (Value : TOrientation);
begin
  TabSet.Orientation := Value;
end;

function tTabbedNotebook95.GetPageIndex : Integer;
begin
  Result := TabSet.TabIndex;
end;

procedure tTabbedNotebook95.SetPageIndex(Value: Integer);
begin
{  Notebook.PageIndex := Value;}
  If (Value>=0) and (Value<=TabSet.Tabs.Count-1)
    then TabSet.TabIndex := Value;
end;

function tTabbedNotebook95.GetPages : tStrings;
begin
  Result := Notebook.Pages;
end;

function tTabbedNotebook95.GetIndexForPage(const PageName: String): Integer;
begin
  Result := Notebook.Pages.IndexOf (PageName);
end;

procedure tTabbedNotebook95.SetPages (Value : tStrings);
begin
{  Notebook.Pages.Clear;
  Notebook.Pages.Assign(Value);}
  Notebook.Pages := Value;
{  TabSet.Tabs.Clear;
  TabSet.Tabs.Assign(Value);}
  TabSet.Tabs := Notebook.Pages;

  Invalidate;
end;

function tTabbedNotebook95.GetTabHeight : integer;
begin
  Result := TabSet.TabHeight;
end;

procedure tTabbedNotebook95.SetTabHeight(Value : integer);
var
  aRect : tRect;
begin
  TabSet.TabHeight := Value;
  aRect := GetClientRect;
  AlignControls(Self, aRect);
end;

function tTabbedNotebook95.GetTabHints : TStrings;
begin
  Result := TabSet.TabHints;
end;

procedure tTabbedNotebook95.SetTabHints(List : TStrings);
begin
  TabSet.TabHints.Clear;
  TabSet.TabHints.Assign(List);
end;

function tTabbedNotebook95.GetTabSpacing : integer;
begin
  Result := TabSet.TabSpacing;
end;

procedure tTabbedNotebook95.SetTabSpacing(Value : integer);
begin
  TabSet.TabSpacing := Value;
end;

procedure tTabbedNotebook95.Paint;
begin
  inherited Paint;
  With canvas do
  begin
    Pen.Width := 1;
    If Assigned (Notebook)
    then begin
      Pen.Color := clWhite;
      Moveto(Notebook.Left-1, Notebook.Top-1);
      LineTo(Notebook.Left-1, Notebook.Top+Notebook.Height+1);
      Pen.Color := clWhite;
      Moveto(Notebook.Left-1, Notebook.Top-1);
      LineTo(Notebook.Left+Notebook.Width+1, Notebook.Top-1);
      Pen.Color := clBtnShadow;
      MoveTo(Notebook.Left-1, Notebook.Top+Notebook.Height);
      LineTo(Notebook.Left+NoteBook.Width, Notebook.Top+Notebook.Height);
      LineTo(Notebook.Left+NoteBook.Width, Notebook.Top-1);
    End;
  end;
end;

procedure tTabbedNotebook95.AlignControls(AControl: TControl; var Rect: TRect);
begin
  try
    Notebook.Top := 1;
    Notebook.Left := 1;
    Notebook.Width := Width-2;
    Notebook.Height := Height-2;
    TabSet.AdjustSize;
    Case TabSet.Orientation of
      toTop : begin
                Notebook.Top := TabSet.Height+1;
                Notebook.Height := Notebook.Height-TabSet.Height;
              end;
      toBottom : begin
                Notebook.Height := Notebook.Height-TabSet.Height;
              end;
      toLeft : begin
                Notebook.Left := TabSet.Width+1;
                Notebook.Width := Notebook.Width-TabSet.Width;
              end;
      toRight : begin
                Notebook.Width := Notebook.Width-TabSet.Width;
              end;
    end;
    TabSet.AdjustSize;
    Invalidate;
  except
    {les conposants ont t dtruits par "Done"}
  end;
  inherited AlignControls (aControl, Rect);
end;

end.


procedure TRotateLabel.DrawAngleText(aCanvas:TCanvas;aRect:tRect;aAngle:Integer;aTxt:String);
{==========================================================================}
{ Draw text with FontIndirect (angle -> escapement)                        }
{--------------------------------------------------------------------------}
var LFont             : TLogFont;
    hOldFont, hNewFont: HFont;
begin
  CalcTextPos(aRect,aAngle,aTxt);

  GetObject(aCanvas.Font.Handle,SizeOf(LFont),Addr(LFont));
  LFont.lfEscapement := aAngle*10;
  hNewFont := CreateFontIndirect(LFont);
  hOldFont := SelectObject(aCanvas.Handle,hNewFont);

  aCanvas.TextOut(aRect.Left,aRect.Top,aTxt);

  hNewFont := SelectObject(aCanvas.Handle,hOldFont);
  DeleteObject(hNewFont);
end;

